/*----------------------------------------------------------------------*\
 | dbfsql.c is a program that reads a DBF file and generates from it	|
 | a set of SQL statements that create a table with the same name as	|
 | the DBF file with fields that match those of the DBF file.  It also	|
 | creates SQL insert statements to load the data from the DBF file		|
 | into the table.														|
 |																		|
 | Based on the dbf access routines developed by Frank Warmerdam as		|
 | part of his shapelib C library for reading and writing shapefiles.	|
 | <http://gdal.velocet.ca/projects/shapelib/>  Thanks, Frank!			|
 |																		|
 | Peter N. Schweitzer (U.S. Geological Survey, Reston VA 20192)		|
\*----------------------------------------------------------------------*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>

#include <unistd.h>

#include "shapefil.h"
#include "stricmp.h"

extern int is_reserved (char *name);

static char version[] = "1.9";

/*----------------------------------------------------------------------*\
 | Revision history:													|
 |																		|
 | 1.9 (20070409) PNS Add option to make the output compatible with		|
 |					  postgresql (no right quotes, '' instead of \',	|
 |					  float8 instead of double)							|
 |																		|
 | 1.8 (20070126) PNS Add option to specify letter case of field and	|
 |					  table names.  Quote all field names rather than	|
 |					  modify those that match reserved words.			|
 |					  Use double for DBF fields of type FTDouble		|
 |					  Use char for string fields of width 4 or less		|
 |																		|
 | 1.7 (20061025) PNS fix SQL insert when last field is to be omitted	|
 |																		|
 | 1.6 (20060914) PNS allow option to substitute a character for ctrls	|
 |																		|
 | 1.5 (20060911) PNS skip list is comma-separated list of fields		|
 |																		|
 | 1.4 (20060215) PNS Add option to skip specified fields				|
 |																		|
 | 1.3 (20030423) PNS Add option to create unique id from row number	|
 |																		|
 | 1.2 (20030422) PNS Add option to specify name of SQL table			|
 |																		|
 | 1.0 (20011119) PNS First release.									|
\*----------------------------------------------------------------------*/

struct field_info {
	char name [16];
	DBFFieldType type;
	int width;
	int precision;
	int keep;
	};

static char *type_of (DBFFieldType t) {
	if (t == FTString)  return ("varchar");
	if (t == FTInteger) return ("int");
	if (t == FTDouble)  return ("double");
	if (t == FTInvalid) return ("invalid");
	return ("unknown");
	}

enum case_t {
	AS_IS = 0,
	LOWER = 1,
	UPPER = 2
	};

enum compatibility_t {
	MYSQL = 0,
	PGSQL = 1,
	ORACLE = 2
	};

static char *empty = "";

#define MAX_TEXT_LENGTH	4096

int main (int argc, char *argv[]) {
	char *input_file = NULL;
	char output_file [FILENAME_MAX];
	char table_name [FILENAME_MAX];
	char text [MAX_TEXT_LENGTH];
	char unique_id [16];
	FILE *out;
	DBFHandle df;
	int rc,fc,max_width;
	int i,j,k;
	DBFFieldType ft[1024];
	char *s,*t;
	struct field_info *field = NULL;
	int do_inserts = 1;
	char *skip_list = NULL;
	char *skip [512];
	enum case_t field_case = AS_IS;
	char what_to_use_for_controls[1] = "\0";
	enum compatibility_t compatibility = MYSQL;

	*table_name = 0;
	*output_file = 0;
	*unique_id = 0;
	skip[0] = NULL;

	if (argc > 1)
		for (i=1; i < argc; i++)
			if (strcmp (argv[i],"-o") == 0) {
				i++;
				if (i < argc) strcpy (output_file,argv[i]);
				}
			else
				if (strcmp (argv[i],"-c") == 0) {
					do_inserts = 0;
					}
				else
					if (strcmp (argv[i],"-t") == 0) {
						i++;
						if (i < argc) strcpy (table_name,argv[i]);
						}
					else
						if (strcmp (argv[i],"-id") == 0) {
							i++;
							if (i < argc) strcpy (unique_id,argv[i]);
							}
						else
							if (strcmp (argv[i],"-ctrl") == 0) {
								i++;
								if (i < argc) *what_to_use_for_controls = *argv[i];
								}
							else
								if (strcmp (argv[i],"-case") == 0) {
									i++;
									if (i < argc) {
										if (stricmp (argv[i],"upper") == 0) field_case = UPPER;
										if (stricmp (argv[i],"lower") == 0) field_case = LOWER;
										}
									}
								else
									if (strcmp (argv[i],"-skip") == 0) {
										i++;
										if (i < argc) {
											skip_list = argv[i];
											}
										}
									else
										if (strcmp (argv[i],"-pg") == 0) {
											compatibility = PGSQL;
											}
										else
											input_file = argv[i];

	if (field_case == UPPER) for (s=table_name; *s; s++) *s = toupper(*s);
	if (field_case == LOWER) for (s=table_name; *s; s++) *s = tolower(*s);

	/*------------------------------------------------------------------*\
	 | The skip_list value might be a comma-separated list of fields	|
	\*------------------------------------------------------------------*/

	if (skip_list) {
		char *b,*e;
		j = 0;
		b = skip_list;
		while (*b) {
			for (e=b; *e && *e != ','; e++);
			if (j < 511) {
				skip[j++] = b;
				skip[j] = NULL;
				}
			b = e;
			if (*b) b++;
			*e = 0;
			}
		}

	printf ("dbfsql %s - Peter N. Schweitzer (U.S. Geological Survey)\n",version);

	if (input_file) {

		/* Open and process the input file */

		if (df = DBFOpen (input_file,"rb")) {

			rc = DBFGetRecordCount (df);
			fc = DBFGetFieldCount (df);

			printf ("Opened file %s\n",input_file);
			printf ("  Number of fields: %d\n",fc);
			printf ("  Number of records: %d\n",rc);

			if (*table_name == 0) {
				/* Figure table name from input file */
				strcpy (table_name,input_file);
				if (s = strrchr (table_name,'.')) *s = 0;
				if (s = strrchr (table_name,'/')) strcpy (table_name,s+1);
				}
			printf ("  Table name will be %s\n",table_name);

			/* Allocate space for the field info */

			if (field = (struct field_info *) malloc (fc * sizeof (struct field_info))) {
				for (i=0; i < fc; i++) {
					field[i].type = DBFGetFieldInfo (df,i,field[i].name,&field[i].width, &field[i].precision);
					field[i].keep = 1;
					}

				/*------------------------------------------------------*\
				 | Look at the names to make sure there aren't any		|
				 | duplicate names.  Warn user if there are.			|
				\*------------------------------------------------------*/

				for (i=0; i < fc - 1; i++)
					for (j=i+1; j < fc; j++)
						if (strcmp (field[i].name,field[j].name) == 0)
							printf ("WARNING: Fields %d and %d have the same name \"%s\"\n",i+1,j+1,field[i].name);
				}

			/*----------------------------------------------------------*\
			 | Mark any fields to be skipped
			\*----------------------------------------------------------*/

			for (i=0; i < fc; i++) {
				for (j=0; skip[j]; j++) {
					if (stricmp (field[i].name,skip[j]) == 0) field[i].keep = 0;
					}
				}

			/*----------------------------------------------------------*\
			 | If the output file is not specified with -o, then make	|
			 | it the same as the input file with .sql instead of .dbf.	|
			\*----------------------------------------------------------*/

			if (!*output_file) {
				strcpy (output_file,input_file);
				if (s = strrchr (output_file,'.')) *s = 0;
				strcat (output_file,".sql");
				}
			printf ("  Output file name: %s\n",output_file);

			if (out = fopen (output_file,"w")) {
				char *q;

				/* Make the SQL CREATE statement */

				switch (compatibility) {
					case MYSQL:
						q = "\x60";
						break;
					case PGSQL:
						q = empty;
						break;
					case ORACLE:
						q = empty;
						break;
					default:
						q = empty;
						break;
					}

				fprintf (out,"create table %s%s%s (\n",q,table_name,q);

				if (*unique_id ) fprintf (out,"\t%s varchar(16),\n",unique_id);

				for (j=0; j < fc; j++)
					if (field[j].keep) {
						char string[16];

						/*--------------------------------------------------*\
						 | Gotta make sure the field name is not reserved	|
						 | by SQL (or in this case by MySQL).				|
						 |													|
						 | I do this by appending an underscore to the name.|
						 | This will always work because the names in the	|
						 | DBF file cannot be more than 10 characters, and	|
						 | I have allocated 16 for the name in the field	|
						 | structure here.  SQL allows longer names.		|
						 |													|
						 | In version 1.8+ This is handled differently, by	|
						 | enclosing all field names in right-quotes.		|
						\*--------------------------------------------------*/

					/*\
						if (is_reserved (field[j].name)) strcat (field[j].name,"_");
					\*/

						sprintf (string,"%s%s%s",q,field[j].name,q);

						switch (field_case) {
							case UPPER:
								for (s=string; *s; s++) *s = toupper(*s);
								break;
							case LOWER:
								for (s=string; *s; s++) *s = tolower(*s);
								break;
							default:
								break;
							}

						fprintf (out,"\t%-13s",string);

						if (field[j].type == FTString && field[j].width < 5)
							fprintf (out," %s","char");
						else
							if (field[j].type == FTDouble && compatibility == PGSQL)
								fprintf (out," %s","float8");
							else
								fprintf (out," %s",type_of(field[j].type));
						if (field[j].type == FTString) fprintf (out,"(%d)",field[j].width);
						if (j < fc - 1) fprintf (out,",");
						fprintf (out,"\n");
						}
				fprintf (out,"\t);\n");

				if (do_inserts) {
					int last_field = 0;

					/*--------------------------------------------------*\
					 | Find the index of the last field that is not to	|
					 | be omitted (that is, not specified with -skip	|
					\*--------------------------------------------------*/

					for (j=0; j < fc; j++)
						if (field[j].keep) last_field = j;

					/*--------------------------------------------------*\
					 | Read the records one by one and build one INSERT	|
					 | statement for each record.						|
					\*--------------------------------------------------*/

					for (i=0; i < rc; i++) {
						fprintf (out,"insert into %s%s%s values (",q,table_name,q);
						if (*unique_id) fprintf (out,"'%d',",i);
						for (j=0; j < fc; j++) {
							s = (char *) DBFReadStringAttribute (df,i,j);
							if (field[j].keep) {
								switch (field[j].type) {

									case FTString:
										if (*s) {
											fprintf (out,"'");
											t = text;
											while (*s)
												if (iscntrl (*s)) {
													if (*what_to_use_for_controls) *t++ = *what_to_use_for_controls;
													s++;
													}
												else {
													if (*s == '\'')
														switch (compatibility) {
															case PGSQL:
															case ORACLE:
																*t++ = '\'';
																break;
															case MYSQL:
															default:
																*t++ = '\\';
																break;
															}
													if (*s == '\\') *t++ = '\\';
													*t++ = *s++;
													}
											*t = 0;
											fprintf (out,"%s",text);
											fprintf (out,"'");
											}
										else
											fprintf (out,"%s","NULL");
										break;

									case FTInteger:
									case FTDouble:
										if (*s == '*' || *s == 0)
											fprintf (out,"%s","NULL");
										else
											fprintf (out,"%s",s);
										break;

									case FTInvalid:
									default:
										fprintf (out,"%s","NULL");
										break;
									}
								if (j < last_field) fprintf (out,",");
								}
							}
						fprintf (out,");\n");
						}
					}

				fclose (out);
				}
			else {
				printf ("Error: could not create output file %s\n",output_file);
				DBFClose (df);
				exit (1);
				}
			DBFClose (df);
			}
		else {
			printf ("Error: could not open input file %s\n",input_file);
			exit (1);
			}
		}
	else {
		printf ("Usage: %s input_file [-o output_file]\n",argv[0]);
		printf ("       -c                show only CREATE statement\n");
		printf ("       -t table_name     specify the SQL table name\n");
		printf ("       -id field_name    create a unique id field\n");
		printf ("       -o output_file    write output to the specified file\n");
		printf ("       -skip name1,name2,... omit the specified fields from the SQL table\n");
		printf ("       -ctrl c           use c instead where any control character appears in the data\n");
		printf ("       -case upper|lower change field names to upper|lower case letters\n");
		printf ("       -pg               create SQL compatible with PostgreSQL\n");
		exit (0);
		}
	exit (0);
	}

/*----------------------------------------------------------------------*\
\*----------------------------------------------------------------------*/
